/*
 * Copyright 2011, Maarten Baert <maarten-baert@hotmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "Buffer.h"

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <new>
#include <map>

#define BUFFER_MINIMUM_CAPACITY 128 // must be a power of two

// shared buffer interface

SharedBufferInterface shared_buffer_iterface;

// implementation of Buffer class

unsigned int buffers_idcounter = 0;
std::map<unsigned int, Buffer*> buffers_list;

Buffer* Buffer::Create() {
	Buffer *buffer = new Buffer();
	buffer->data = NULL;
	buffer->length = 0;
	buffer->capacity = 0;
	buffer->pos = 0;
	buffer->error = false;
	buffer->id = ++buffers_idcounter;
	buffers_list.insert(std::pair<unsigned int, Buffer*>(buffer->id, buffer));
	return buffer;
}

void Buffer::Destroy(Buffer* buffer) {
	buffers_list.erase(buffer->id);
	free(buffer->data);
	delete buffer;
}

unsigned int Buffer::GetID(Buffer* buffer) {
	return buffer->id;
}

Buffer* Buffer::Find(unsigned int id) {
	std::map<unsigned int, Buffer*>::iterator it = buffers_list.find(id);
	return (it == buffers_list.end())? NULL : it->second;
}

const char* Buffer::ToString(Buffer* buffer) {
	return (buffer->data == NULL)? "" : buffer->data;
}

char* Buffer::GetData(Buffer* buffer) {
	return buffer->data;
}

unsigned int Buffer::GetPos(Buffer* buffer) {
	return buffer->pos;
}

unsigned int Buffer::GetLength(Buffer* buffer) {
	return buffer->length;
}

bool Buffer::IsAtEnd(Buffer* buffer) {
	return (buffer->pos == buffer->length);
}

bool Buffer::GetError(Buffer* buffer) {
	return buffer->error;
}

void Buffer::ClearError(Buffer* buffer) {
	buffer->error = false;
}

void Buffer::Clear(Buffer* buffer) {
	free(buffer->data);
	buffer->data = NULL;
	buffer->length = 0;
	buffer->capacity = 0;
	buffer->pos = 0;
	buffer->error = false;
}

void Buffer::SetPos(Buffer* buffer, unsigned int newpos) {
	buffer->pos = (newpos > buffer->length)? buffer->length : newpos;
}

void Buffer::SetLength(Buffer* buffer, unsigned int newlength) {
	
	// The memory block is not resized every time because this is slow.
	// The capacity is always a power of two.
	if(newlength >= buffer->capacity || newlength + BUFFER_MINIMUM_CAPACITY / 2 < buffer->capacity / 2) {
		
		// calculate the next power of two
		unsigned int newcapacity = newlength | (BUFFER_MINIMUM_CAPACITY - 1);
		newcapacity |= newcapacity >> 1;
		newcapacity |= newcapacity >> 2;
		newcapacity |= newcapacity >> 4;
		newcapacity |= newcapacity >> 8;
		newcapacity |= newcapacity >> 16;
		++newcapacity;
		
		// allocate memory
		void *temp = realloc(buffer->data, newcapacity);
		if(temp == NULL) {
			throw std::bad_alloc();
		}
		buffer->data = (char*)(temp);
		buffer->capacity = newcapacity;
		
	}
	
	// change position and length
	buffer->data[newlength] = '\0';
	if(buffer->pos > newlength) buffer->pos = newlength;
	buffer->length = newlength;
	
}

bool Buffer::ReadFromFile(Buffer* buffer, const char* filename) {
	Clear(buffer);
	FILE *f = fopen(filename, "rb");
	if(f == NULL) return false;
	fseek(f, 0, SEEK_END);
	long size = ftell(f);
	fseek(f, 0, SEEK_SET);
	if(size < 0l) {
		fclose(f);
		return false;
	}
	try {
		SetLength(buffer, size);
	} catch(...) {
		fclose(f);
		throw;
	}
	if(fread(buffer->data, 1, size, f) != (size_t)(size)) {
		Clear(buffer);
		fclose(f);
		return false;
	}
	fclose(f);
	return true;
}

bool Buffer::WriteToFile(Buffer* buffer, const char* filename) {
	FILE *f = fopen(filename, "wb");
	if(f == NULL) return false;
	if(fwrite(buffer->data, 1, buffer->length, f) != buffer->length) {
		fclose(f);
		return false;
	}
	fclose(f);
	return true;
}

template<typename X>
X Buffer::ReadX(Buffer* buffer) {
	if(buffer->error) return X();
	if(sizeof(X) > buffer->length - buffer->pos) { // overflow-safe
		buffer->error = true;
		return X();
	}
	X *value = (X*)(buffer->data + buffer->pos);
	buffer->pos += sizeof(X);
	return *value;	
}

template<typename X>
void Buffer::WriteX(Buffer* buffer, X value) {
	if(buffer->pos + sizeof(X) > buffer->length) {
		SetLength(buffer, buffer->pos + sizeof(X));
	}
	*((X*)(buffer->data + buffer->pos)) = value;
	buffer->pos += sizeof(X);
}

const char* Buffer::ReadString(Buffer* buffer) {
	if(buffer->error) return "";
	for(unsigned int i = buffer->pos; i < buffer->length; ++i) {
		if(buffer->data[i] == '\0') {
			const char *str = buffer->data + buffer->pos;
			buffer->pos = i + 1;
			return str;
		}
	}
	buffer->error = true;
	return "";
}

void Buffer::WriteString(Buffer* buffer, const char* string) {
	unsigned int bytes = strlen(string) + 1;
	if(buffer->pos + bytes > buffer->length) {
		SetLength(buffer, buffer->pos + bytes);
	}
	memcpy(buffer->data + buffer->pos, string, bytes);
	buffer->pos += bytes;
}

void Buffer::ReadData(Buffer* buffer, char* ptr, unsigned int bytes) {
	if(buffer->error) {
		memset(ptr, 0, bytes);
		return;
	}
	if(bytes > buffer->length - buffer->pos) { // overflow-safe
		buffer->error = true;
		memset(ptr, 0, bytes);
		return;
	}
	if(bytes != 0) {
		memcpy(ptr, buffer->data + buffer->pos, bytes);
		buffer->pos += bytes;
	}
}

void Buffer::WriteData(Buffer* buffer, const char* ptr, unsigned int bytes) {
	if(buffer->pos + bytes > buffer->length) {
		SetLength(buffer, buffer->pos + bytes);
	}
	memcpy(buffer->data + buffer->pos, ptr, bytes);
	buffer->pos += bytes;
}

void bintohex(char* output, const char* input, unsigned int bytes) {
	for(unsigned int i = 0; i < bytes; ++i) {
		unsigned char a = input[i];
		output[2 * i]     = ((a >> 4) > 9)? 'a' + (a >> 4) - 10 : '0' + (a >> 4);
		output[2 * i + 1] = ((a & 15) > 9)? 'a' + (a & 15) - 10 : '0' + (a & 15);
	}
}

void hextobin(char* output, const char* input, unsigned int bytes) {
	for(unsigned int i = 0; i < bytes; ++i) {
		unsigned char a = input[2*i];
		unsigned char b = input[2*i+1];
		output[i] =
			((('0' <= a && a <= '9')? a - '0' : (('a' <= a && a <= 'f')? 10 + a - 'a' : (('A' <= a && a <= 'F')? 10 + a - 'A' : 0))) << 4) |
			(('0' <= b && b <= '9')? b - '0' : (('a' <= b && b <= 'f')? 10 + b - 'a' : (('A' <= b && b <= 'F')? 10 + b - 'A' : 0)));
	}
}

void Buffer::ReadHex(Buffer* buffer, char* ptr, unsigned int bytes) {
	if(buffer->error) {
		memset(ptr, '0', bytes * 2);
		return;
	}
	if(bytes > buffer->length - buffer->pos) { // overflow-safe
		buffer->error = true;
		memset(ptr, '0', bytes * 2);
		return;
	}
	if(bytes != 0) {
		bintohex(ptr, buffer->data + buffer->pos, bytes);
		buffer->pos += bytes;
	}
}

void Buffer::WriteHex(Buffer* buffer, const char* ptr, unsigned int bytes) {
	if(buffer->pos + bytes > buffer->length) {
		SetLength(buffer, buffer->pos + bytes);
	}
	hextobin(buffer->data + buffer->pos, ptr, bytes);
	buffer->pos += bytes;
}

void Buffer::WriteBuffer(Buffer* buffer, Buffer* source) {
	unsigned int bytes = source->length; // save the old length in case buffer == source
	if(buffer->pos + bytes > buffer->length) {
		SetLength(buffer, buffer->pos + bytes);
	}
	memmove(buffer->data + buffer->pos, source->data, bytes);
	buffer->pos += bytes;
}
void Buffer::WriteBufferPart(Buffer* buffer, Buffer* source, unsigned int offset, unsigned int bytes) {
	if(offset > source->length) offset = source->length;
	if(bytes > source->length - offset) bytes = source->length - offset; // overflow-safe
	if(buffer->pos + bytes > buffer->length) {
		SetLength(buffer, buffer->pos + bytes);
	}
	memmove(buffer->data + buffer->pos, source->data + offset, bytes);
	buffer->pos += bytes;
}

// dummy class to initialize the shared buffer interface
// and destroy buffers when the DLL is unloaded
class InitializeSharedBufferInterface {
	public:
	InitializeSharedBufferInterface() {
		shared_buffer_iterface.create = &Buffer::Create;
		shared_buffer_iterface.destroy = &Buffer::Destroy;
		shared_buffer_iterface.getid = &Buffer::GetID;
		shared_buffer_iterface.find = &Buffer::Find;
		shared_buffer_iterface.tostring = &Buffer::ToString;
		shared_buffer_iterface.getdata = &Buffer::GetData;
		shared_buffer_iterface.getpos = &Buffer::GetPos;
		shared_buffer_iterface.getlength = &Buffer::GetLength;
		shared_buffer_iterface.isatend = &Buffer::IsAtEnd;
		shared_buffer_iterface.geterror = &Buffer::GetError;
		shared_buffer_iterface.clearerror = &Buffer::ClearError;
		shared_buffer_iterface.clear = &Buffer::Clear;
		shared_buffer_iterface.setpos = &Buffer::SetPos;
		shared_buffer_iterface.setlength = &Buffer::SetLength;
		shared_buffer_iterface.readfromfile = &Buffer::ReadFromFile;
		shared_buffer_iterface.writetofile = &Buffer::WriteToFile;
		shared_buffer_iterface.readint8 = &Buffer::ReadX<int8_t>;
		shared_buffer_iterface.readuint8 = &Buffer::ReadX<uint8_t>;
		shared_buffer_iterface.readint16 = &Buffer::ReadX<int16_t>;
		shared_buffer_iterface.readuint16 = &Buffer::ReadX<uint16_t>;
		shared_buffer_iterface.readint32 = &Buffer::ReadX<int32_t>;
		shared_buffer_iterface.readuint32 = &Buffer::ReadX<uint32_t>;
		shared_buffer_iterface.readint64 = &Buffer::ReadX<int64_t>;
		shared_buffer_iterface.readuint64 = &Buffer::ReadX<uint64_t>;
		shared_buffer_iterface.readfloat32 = &Buffer::ReadX<float>;
		shared_buffer_iterface.readfloat64 = &Buffer::ReadX<double>;
		shared_buffer_iterface.writeint8 = &Buffer::WriteX<int8_t>;
		shared_buffer_iterface.writeuint8 = &Buffer::WriteX<uint8_t>;
		shared_buffer_iterface.writeint16 = &Buffer::WriteX<int16_t>;
		shared_buffer_iterface.writeuint16 = &Buffer::WriteX<uint16_t>;
		shared_buffer_iterface.writeint32 = &Buffer::WriteX<int32_t>;
		shared_buffer_iterface.writeuint32 = &Buffer::WriteX<uint32_t>;
		shared_buffer_iterface.writeint64 = &Buffer::WriteX<int64_t>;
		shared_buffer_iterface.writeuint64 = &Buffer::WriteX<uint64_t>;
		shared_buffer_iterface.writefloat32 = &Buffer::WriteX<float>;
		shared_buffer_iterface.writefloat64 = &Buffer::WriteX<double>;
		shared_buffer_iterface.readstring = &Buffer::ReadString;
		shared_buffer_iterface.writestring = &Buffer::WriteString;
		shared_buffer_iterface.readdata = &Buffer::ReadData;
		shared_buffer_iterface.writedata = &Buffer::WriteData;
		shared_buffer_iterface.readhex = &Buffer::ReadHex;
		shared_buffer_iterface.writehex = &Buffer::WriteHex;
		shared_buffer_iterface.writebuffer = &Buffer::WriteBuffer;
		shared_buffer_iterface.writebufferpart = &Buffer::WriteBufferPart;
	}
	~InitializeSharedBufferInterface() {
		for(std::map<unsigned int, Buffer*>::iterator it = buffers_list.begin(); it != buffers_list.end(); it++) {
			delete it->second;
		}
	}
} initialize_shared_buffer_interface;

